import datetime
import inspect
import os
import re
import unittest

import numpy

import cf

class FieldTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        self.filename2 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      'test_file2.nc')
        self.f = cf.read(self.filename)

        self.test_only = []
#        self.test_only = ['test_FieldList_select_by_construct']


    def test_FieldList(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)[0]
        g = cf.FieldList(f)


    def test_FieldList__add__iadd__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)

        f = f + f.copy()
        self.assertTrue(len(f) == 2)
        self.assertIsInstance(f, cf.FieldList)

        f += (f[0].copy(),)
        self.assertTrue(len(f) == 3)

        f += [f[0].copy()]
        self.assertTrue(len(f) == 4)
        self.assertIsInstance(f, cf.FieldList)

        f += list(f[0].copy())
        self.assertTrue(len(f) == 5)

        f += f.copy()
        self.assertTrue(len(f) == 10)

        f = f + f.copy()
        self.assertTrue(len(f) == 20)


    def test_FieldList__contains__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        f.append(f[0].copy())
        f[1] *= 10
        g = cf.read(self.filename)[0] * 10
        self.assertTrue(g in f)
        self.assertFalse(34.6 in f)


    def test_FieldList_close(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        self.assertTrue(f.close() is None)

        _ = repr(f[0])


    def test_FieldList__len__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)

        self.assertTrue(len(cf.FieldList()) == 0)
        self.assertTrue(len(f) == 1)
        f.append(f[0].copy())
        self.assertTrue(len(f) == 2)
        f.extend(f.copy())
        self.assertTrue(len(f) == 4)


    def test_FieldList__mul__imul__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.FieldList()
        f = f * 4
        self.assertTrue(len(f) == 0)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.FieldList()
        f *= 4
        self.assertTrue(len(f) == 0)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.read(self.filename)
        f = f * 4
        self.assertTrue(len(f) == 4)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.read(self.filename)
        f *= 4
        self.assertTrue(len(f) == 4)
        self.assertIsInstance(f, cf.FieldList)

        f = f * 2
        self.assertTrue(len(f) == 8)
        self.assertIsInstance(f, cf.FieldList)

        f *= 3
        self.assertTrue(len(f) == 24)
        self.assertIsInstance(f, cf.FieldList)


    def test_FieldList__repr__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        f += f

        _ = repr(f)


    def test_FieldList_append_extend(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Append
        f = cf.FieldList()

        f.append(cf.read(self.filename)[0])
        self.assertTrue(len(f) == 1)
        self.assertIsInstance(f, cf.FieldList)

        f.append(f[0].copy())
        self.assertTrue(len(f) == 2)
        self.assertIsInstance(f, cf.FieldList)

        f.append(f[0].copy())
        self.assertTrue(len(f) == 3)

        # Extend
        f = cf.FieldList()

        f.extend(cf.read(self.filename))
        self.assertTrue(len(f) == 1)
        self.assertIsInstance(f, cf.FieldList)

        f.extend(f.copy())
        self.assertTrue(len(f) == 2)
        self.assertIsInstance(f, cf.FieldList)

        f.extend(f.copy())
        self.assertTrue(len(f) == 4)


    def test_FieldList_copy(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        f.append(f[0].copy())
        g = f.copy()
        self.assertTrue(f.equals(f, verbose=True))
        self.assertTrue(f.equals(g, verbose=True))


    def test_FieldList__getslice__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        f.append(f[0].copy())

        _ = f[0:1]
        _ = f[1:2]
        _ = f[:1]
        _ = f[1:]


    def test_FieldList_count(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)

        self.assertTrue(f.count(f[0]) == 1)

        f *= 7
        self.assertTrue(f.count(f[0]) == 7)

        f[3] = f[0] * 99
        f[5] = f[0] * 99
        self.assertTrue(f.count(f[0]) == 5)
        self.assertTrue(f.count(f[3]) == 2)


    def test_FieldList_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        g = f.copy()
        self.assertTrue(f.equals(f, verbose=True))
        self.assertTrue(f.equals(g, verbose=True))

        f += g.copy()
        self.assertTrue(f.equals(f, verbose=True))
        self.assertTrue(len(f) == 2)
        g = f.copy()
        self.assertTrue(f.equals(g, verbose=True))
        self.assertTrue(f.equals(g, unordered=True, verbose=True))

        h = cf.read(self.filename2)
        f.extend(h)
        self.assertTrue(f.equals(f, verbose=True))
        self.assertTrue(f.equals(f.copy(), verbose=True))

        g = f.copy()[::-1]
        self.assertFalse(f.equals(g))
        self.assertTrue(f.equals(g, unordered=True, verbose=True))

        g = g[:-1]
        self.assertFalse(f.equals(g))
        self.assertFalse(f.equals(g, unordered=True))

        g.append(g[0])
        self.assertFalse(f.equals(g))
        self.assertFalse(f.equals(g, unordered=True))

        h *= 3
        self.assertFalse(f.equals(h))
        self.assertFalse(f.equals(h, unordered=True))


    def test_FieldList_insert_pop_remove(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Insert
        f = cf.read(self.filename)
        g = f[0].copy()

        f.insert(0, g.copy())
        self.assertTrue(len(f) == 2)
        self.assertIsInstance(f, cf.FieldList)

        g = g + 10
        f.insert(-1, g)
        self.assertTrue(len(f) == 3)
        self.assertTrue(f[0].maximum() == (f[1].maximum() - 10))
        self.assertTrue(isinstance(f, cf.FieldList))

        # Pop
        f = cf.read(self.filename)
        g = f[0]
        h = f[0] + 10
        f.append(h)

        z = f.pop(0)
        self.assertTrue(z is g)
        self.assertTrue(len(f) == 1)
        self.assertIsInstance(f, cf.FieldList)

        z = f.pop(-1)
        self.assertTrue(z is h)
        self.assertTrue(len(f) == 0)
        self.assertIsInstance(f, cf.FieldList)

        # Remove
        f = cf.read(self.filename)
        g = f[0].copy()
        g = g + 10

        f.append(g)
        self.assertTrue(len(f) == 2)

        f.remove(g)
        self.assertTrue(len(f) == 1)
        self.assertIsInstance(f, cf.FieldList)

        with self.assertRaises(Exception):
            f.remove(f[0]*-99)

        f.remove(f[0].copy())
        self.assertTrue(len(f) == 0)
        self.assertIsInstance(f, cf.FieldList)


    def test_FieldList_reverse(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)
        g = f[0]
        h = f[0] + 10
        f.append(h)

        self.assertTrue(g is f[0])
        self.assertTrue(h is f[1])

        f.reverse()
        self.assertIsInstance(f, cf.FieldList)
        self.assertTrue(len(f) == 2)
        self.assertTrue(g is f[1])
        self.assertTrue(h is f[0])


    def test_FieldList_select(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)

        g = f('not this one')
        self.assertIsInstance(g, cf.FieldList)
        self.assertTrue(len(g) == 0)

        g = f('eastward_wind')
        self.assertIsInstance(g, cf.FieldList)
        self.assertTrue(len(g) == 1, len(g))

        g = f(re.compile('^eastw'))
        self.assertIsInstance(g, cf.FieldList)
        self.assertTrue(len(g) == 1, len(g))

        f *= 9
        f[4] = f[0].copy()
        f[4].standard_name = 'this one'
        f[6] = f[0].copy()
        f[6].standard_name = 'this one'

        g = f(re.compile('^eastw'))
        self.assertIsInstance(g, cf.FieldList)
        self.assertTrue(len(g) == 7, len(g))

        g = f('this one')
        self.assertIsInstance(g, cf.FieldList)
        self.assertTrue(len(g) == 2)

        # select_by_Units
        f[1] = f[1].override_units(cf.Units('watt'))
        f[3] = f[3].override_units(cf.Units('mile hour-1'))

        self.assertTrue(len(f.select_by_units()) == 9)
        self.assertTrue(len(f.select_by_units(cf.Units('m s-1'))) == 7)
        self.assertTrue(len(f.select_by_units(cf.Units('m s-1'), exact=False)) == 8)
        self.assertTrue(len(f.select_by_units('m s-1')) == 7)
        self.assertTrue(len(f.select_by_units('m s-1', exact=False)) == 8)
        self.assertTrue(len(f.select_by_units(re.compile('^mile|watt'))) == 2)

        # select_by_Units
        f[1] = f[1].override_units(cf.Units('watt'))
        f[3] = f[3].override_units(cf.Units('mile hour-1'))

        self.assertTrue(len(f.select_by_units()) == 9)
        self.assertTrue(len(f.select_by_units(cf.Units('m s-1'))) == 7)
        self.assertTrue(len(f.select_by_units(cf.Units('m s-1'), exact=False)) == 8)
        self.assertTrue(len(f.select_by_units('m s-1')) == 7)
        self.assertTrue(len(f.select_by_units('m s-1', exact=False)) == 8)
        self.assertTrue(len(f.select_by_units(re.compile('^mile|watt'))) == 2)

        self.assertTrue(len(f.select_by_units('long_name=qwery:asd')) == 0)


    def test_FieldList_select_by_construct(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read([self.filename, self.filename2])

        g = f.select_by_construct()
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('latitude')
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('latitude', 'longitude')
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('latitude', 'time')
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct('latitude', 'time', OR=False)
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct('latitude', 'time', OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct(longitude=cf.gt(0))
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct(longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')))
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct(longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('longitude', longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('latitude', longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('time', longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('time', longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=False)
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct('time', longitude=cf.gt(0), time=cf.le(cf.dt('2008-12-01')), OR=False)
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct('qwerty')
        self.assertTrue(len(g) == 0)

        g = f.select_by_construct('qwerty', 'latitude')
        self.assertTrue(len(g) == 0)

        g = f.select_by_construct('qwerty', 'latitude', OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct('qwerty', 'time', 'longitude')
        self.assertTrue(len(g) == 0)

        g = f.select_by_construct('qwerty', 'time', 'longitude', OR=True)
        self.assertTrue(len(g) == 2)

        g = f.select_by_construct(longitude=cf.gt(70))
        self.assertTrue(len(g) == 1)

        g = f.select_by_construct(longitude=cf.gt(0), time=cf.le(cf.dt('1999-12-01')))
        self.assertTrue(len(g) == 0)


    def test_FieldList_select_field(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename)

        with self.assertRaises(Exception):
            _ = f.select_field('not this one')

        self.assertTrue(f.select_field('not this one', None) is None)

        g = f.select_field('eastward_wind')
        self.assertIsInstance(g, cf.Field)

        g = f.select_field(re.compile('^eastw'))
        self.assertIsInstance(g, cf.Field)

        with self.assertRaises(Exception):
            g = f.select_field(re.compile('^QWERTY'))

    def test_FieldList_concatenate(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename2)[0]

        g = cf.FieldList([f[0], f[1:456], f[456:]])

        h = g.concatenate(axis=0)
        self.assertTrue(f.equals(h, verbose=True))

        h = g.concatenate(axis=0, _preserve=False)
        self.assertTrue(f.equals(h, verbose=True))

    def test_FieldList_index(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename2)[0]

        a, b, c = [f[0], f[1:456], f[456:]]
        g = cf.FieldList([a, b, c])

        self.assertTrue(g.index(a) == 0)
        self.assertTrue(g.index(a, start=0) == 0)
        self.assertTrue(g.index(a, stop=1) == 0)
        self.assertTrue(g.index(a, stop=-2) == 0)
        self.assertTrue(g.index(a, stop=2) == 0)
        self.assertTrue(g.index(b) == 1)
        self.assertTrue(g.index(b, start=0) == 1)
        self.assertTrue(g.index(b, start=1, stop=2) == 1)
        self.assertTrue(g.index(c) == 2)
        self.assertTrue(g.index(c, start=0) == 2)
        self.assertTrue(g.index(c, start=1) == 2)
        self.assertTrue(g.index(c, start=2) == 2)
        self.assertTrue(g.index(c, start=-1) == 2)

        with self.assertRaises(Exception):
            _ = g.index(f)

        with self.assertRaises(Exception):
            _ = g.index(a, start=1)


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import unittest
import inspect
import subprocess

import cf


class cfaTest(unittest.TestCase):
    def setUp(self):
        self.test_only = ()

    def test_cfa(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        subprocess.run(' '.join(['.', './cfa_test.sh']),
                       shell=True, check=True)


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
from numpy import array as numpy_array
from numpy import empty as numpy_empty
from numpy import where as numpy_where
from numpy import sum as numpy_sum
from numpy import finfo as numpy_finfo
from .data.data import Data
from .dimensioncoordinate import DimensionCoordinate
from .functions import REGRID_LOGGING
from . import _found_ESMF
if _found_ESMF:
    try:
        import ESMF
    except Exception as error:
        print("WARNING: Can not import ESMF for regridding: {0}".format(error))


class Regrid:
    '''Class containing all the methods required for accessing ESMF
    regridding through ESMPY and the associated utility methods.

    '''

    def __init__(self, srcfield, dstfield, srcfracfield, dstfracfield,
                 method='conservative_1st', ignore_degenerate=False):
        '''Creates a handle for regridding fields from a source grid to a
    destination grid that can then be used by the run_regridding method.

    :Parameters:

        srcfield: ESMF.Field
            The source field with an associated grid to be used for
            regridding.

        dstfield: ESMF.Field
            The destination field with an associated grid to be used
            for regridding.

        srcfracfield: ESMF.Field
            A field to hold the fraction of the source field that
            contributes to conservative regridding.

        dstfracfield: ESMF.Field
            A field to hold the fraction of the source field that
            contributes to conservative regridding.

        method: `str`, optional
            By default the regridding method is set to
            'conservative_1st'. In this case or if it is set to
            'conservative' first-order conservative regridding is
            used. If it is set to 'conservative_2nd' second order
            conservative regridding is used. If it is set to
            'linear' then (multi)linear interpolation is used.  If
            it is set to 'patch' then higher-order patch recovery is
            used.  If it is set to 'nearest_stod' then nearest source
            to destination interpolation is used. If it is set to
            'nearest_dtos' then nearest destination to source
            interpolation is used.

        ignore_degenerate: `bool`, optional
            Whether to check for degenerate points.

        '''
        # create a handle to the regridding method
        regrid_method_map = {
            'linear': ESMF.RegridMethod.BILINEAR,  # see comment below...
            'bilinear': ESMF.RegridMethod.BILINEAR,  # (for back compat)
            'conservative': ESMF.RegridMethod.CONSERVE,
            'conservative_1st': ESMF.RegridMethod.CONSERVE,
            'conservative_2nd': ESMF.RegridMethod.CONSERVE_2ND,
            'nearest_dtos': ESMF.RegridMethod.NEAREST_DTOS,
            'nearest_stod': ESMF.RegridMethod.NEAREST_STOD,
            'patch': ESMF.RegridMethod.PATCH,
        }
        # ... diverge from ESMF with respect to name for bilinear method by
        # using 'linear' because 'bi' implies 2D linear interpolation, which
        # could mislead or confuse for Cartesian regridding in 1D or 3D.
        regrid_method = regrid_method_map.get(
            method, ValueError('Regrid method not recognised.'))
        # Initialise the regridder. This also creates the
        # weights needed for the regridding.
        self.regridSrc2Dst = ESMF.Regrid(
            srcfield, dstfield, regrid_method=regrid_method,
            src_mask_values=numpy_array([0], dtype='int32'),
            dst_mask_values=numpy_array([0], dtype='int32'),
            src_frac_field=srcfracfield, dst_frac_field=dstfracfield,
            unmapped_action=ESMF.UnmappedAction.IGNORE,
            ignore_degenerate=ignore_degenerate)

    def destroy(self):
        '''Free the memory associated with the ESMF.Regrid instance.

        '''
        self.regridSrc2Dst.destroy()

    @staticmethod
    def initialize():
        '''Check whether ESMF has been found. If not raise an import
    error. Initialise the ESMPy manager. Whether logging is enabled or
    not is determined by cf.REGRID_LOGGING. If it is then logging
    takes place after every call to ESMPy.

    :Returns:

        `ESMF.Manager`
            A singleton instance of the ESMPy manager.

        '''
        if not _found_ESMF:
            raise ImportError(
                'The ESMF package is needed to support regridding.')

        manager = ESMF.Manager(debug=REGRID_LOGGING())

        return manager

    @staticmethod
    def create_grid(coords, use_bounds, mask=None, cartesian=False,
                    cyclic=False, coords_2D=False, coord_order=None):
        '''Create an ESMPy grid given a sequence of coordinates for use as a
    source or destination grid in regridding. Optionally the grid may
    have an associated mask.

    :Parameters:

        coords: sequence
            The coordinates if not Cartesian it is assume that the
            first is longitude and the second is latitude.

        use_bounds: `bool`
            Whether to populate the grid corners with information from
            the bounds or not.

        mask: `numpy.ndarray`, optional

            An optional numpy array of booleans containing the grid
            points to mask.  Where the elements of mask are True the
            output grid is masked.

        cartesian: `bool`, optional
            Whether to create a Cartesian grid or a spherical one,
            False by default.

        cyclic: `bool`, optional
            Whether or not the longitude (if present) is cyclic. If
            None the a check for cyclicity is made from the
            bounds. None by default.

        coords_2D: `bool`, optional
            Whether the coordinates are 2D or not. Presently only
            works for spherical coordinates. False by default.

        coord_order: sequence, optional
            Two tuples one indicating the order of the x and y axes
            for 2D longitude, one for 2D latitude.

    :Returns:

        `ESMF.Grid`
            The resulting ESMPy grid for use as a source or destination
            grid in regridding.

        '''

        if not cartesian:
            lon = coords[0]
            lat = coords[1]
            if not coords_2D:
                # Get the shape of the grid
                shape = [lon.size, lat.size]
            else:
                x_order = coord_order[0]
                y_order = coord_order[1]
                # Get the shape of the grid
                shape = lon.transpose(x_order).shape
                if lat.transpose(y_order).shape != shape:
                    raise ValueError(
                        'The longitude and latitude coordinates'
                        ' must have the same shape.'
                    )
            # --- End: if

            if use_bounds:
                if not coords_2D:
                    # Get the bounds
                    x_bounds = lon.get_bounds()
                    y_bounds = lat.get_bounds().clip(-90, 90, 'degrees').array

                    # If cyclic not set already, check for cyclicity
                    if cyclic is None:
                        cyclic = abs(x_bounds.datum(-1)
                                     - x_bounds.datum(0)) == Data(360,
                                                                  'degrees')

                    x_bounds = x_bounds.array
                else:
                    # Get the bounds
                    x_bounds = lon.get_bounds()
                    y_bounds = lat.get_bounds().clip(-90, 90, 'degrees')
                    n = x_bounds.shape[0]
                    m = x_bounds.shape[1]
                    x_bounds = x_bounds.array
                    y_bounds = y_bounds.array

                    tmp_x = numpy_empty((n + 1, m + 1))
                    tmp_x[:n, :m] = x_bounds[:, :, 0]
                    tmp_x[:n, m] = x_bounds[:, -1, 1]
                    tmp_x[n, :m] = x_bounds[-1, :, 3]
                    tmp_x[n, m] = x_bounds[-1, -1, 2]

                    tmp_y = numpy_empty((n + 1, m + 1))
                    tmp_y[:n, :m] = y_bounds[:, :, 0]
                    tmp_y[:n, m] = y_bounds[:, -1, 1]
                    tmp_y[n, :m] = y_bounds[-1, :, 3]
                    tmp_y[n, m] = y_bounds[-1, -1, 2]

                    x_bounds = tmp_x
                    y_bounds = tmp_y

            else:
                if not coords_2D:
                    # If cyclicity not set already, check for cyclicity
                    if cyclic is None:
                        try:
                            x_bounds = lon.get_bounds()
                            cyclic = abs(x_bounds.datum(-1) -
                                         x_bounds.datum(0)) == Data(
                                             360, 'degrees')
                        except ValueError:
                            pass
            # --- End: if

            # Create empty grid
            max_index = numpy_array(shape, dtype='int32')
            if use_bounds:
                staggerLocs = [ESMF.StaggerLoc.CORNER, ESMF.StaggerLoc.CENTER]
            else:
                staggerLocs = [ESMF.StaggerLoc.CENTER]

            if cyclic:
                grid = ESMF.Grid(max_index, num_peri_dims=1,
                                 staggerloc=staggerLocs)
            else:
                grid = ESMF.Grid(max_index, staggerloc=staggerLocs)

            # Populate grid centres
            x, y = 0, 1
            gridXCentre = grid.get_coords(x, staggerloc=ESMF.StaggerLoc.CENTER)
            gridYCentre = grid.get_coords(y, staggerloc=ESMF.StaggerLoc.CENTER)
            if not coords_2D:
                gridXCentre[...] = lon.array.reshape((lon.size, 1))
                gridYCentre[...] = lat.array.reshape((1, lat.size))
            else:
                gridXCentre[...] = lon.transpose(x_order).array
                gridYCentre[...] = lat.transpose(y_order).array

            # Populate grid corners if there are bounds
            if use_bounds:
                gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
                if not coords_2D:
                    if cyclic:
                        gridCorner[x][...] = x_bounds[:, 0].reshape(
                            lon.size, 1)
                    else:
                        n = x_bounds.shape[0]
                        tmp_x = numpy_empty(n + 1)
                        tmp_x[:n] = x_bounds[:, 0]
                        tmp_x[n] = x_bounds[-1, 1]
                        gridCorner[x][...] = tmp_x.reshape(lon.size + 1, 1)

                    n = y_bounds.shape[0]
                    tmp_y = numpy_empty(n + 1)
                    tmp_y[:n] = y_bounds[:, 0]
                    tmp_y[n] = y_bounds[-1, 1]
                    gridCorner[y][...] = tmp_y.reshape(1, lat.size + 1)
                else:
                    gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
                    x_bounds = x_bounds.transpose(x_order)
                    y_bounds = y_bounds.transpose(y_order)
                    if cyclic:
                        x_bounds = x_bounds[:-1, :]
                        y_bounds = y_bounds[:-1, :]
                    gridCorner[x][...] = x_bounds
                    gridCorner[y][...] = y_bounds
            # --- End: if
        else:
            # Test the dimensionality of the list of coordinates
            ndim = len(coords)
            if ndim < 1 or ndim > 3:
                raise ValueError(
                    'Cartesian grid must have between 1 and 3 dimensions.'
                )

            # For 1D conservative regridding add an extra dimension of size 1
            if ndim == 1:
                if not use_bounds:
                    # For 1D nonconservative regridding the extra dimension
                    # should already have been added in cf.Field.regridc.
                    raise ValueError(
                        'Cannot create a Cartesian grid from '
                        'one dimension coordinate with no bounds.'
                    )
                coords = [DimensionCoordinate(data=Data(0),
                          bounds=Data([numpy_finfo('float32').epsneg,
                                       numpy_finfo('float32').eps]))] + coords
                if mask is not None:
                    mask = mask[None, :]
                ndim = 2

            shape = list()
            for coord in coords:
                shape.append(coord.size)

            # Initialise the grid
            max_index = numpy_array(shape, dtype='int32')
            if use_bounds:
                if ndim < 3:
                    staggerLocs = [ESMF.StaggerLoc.CORNER,
                                   ESMF.StaggerLoc.CENTER]
                else:
                    staggerLocs = [ESMF.StaggerLoc.CENTER_VCENTER,
                                   ESMF.StaggerLoc.CORNER_VFACE]
            else:
                if ndim < 3:
                    staggerLocs = [ESMF.StaggerLoc.CENTER]
                else:
                    staggerLocs = [ESMF.StaggerLoc.CENTER_VCENTER]
            # --- End: if
            grid = ESMF.Grid(max_index, coord_sys=ESMF.CoordSys.CART,
                             staggerloc=staggerLocs)

            # Populate the grid centres
            for d in range(0, ndim):
                if ndim < 3:
                    gridCentre = grid.get_coords(
                        d, staggerloc=ESMF.StaggerLoc.CENTER)
                else:
                    gridCentre = grid.get_coords(
                        d, staggerloc=ESMF.StaggerLoc.CENTER_VCENTER)
                gridCentre[...] = coords[d].array.reshape(
                    [shape[d] if x == d else 1 for x in range(0, ndim)])
            # --- End: for

            # Populate grid corners
            if use_bounds:
                if ndim < 3:
                    gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
                else:
                    gridCorner = grid.coords[ESMF.StaggerLoc.CORNER_VFACE]

                for d in range(0, ndim):
                    # boundsD = coords[d].get_bounds(create=True).array
                    boundsD = coords[d].get_bounds(None)
                    if boundsD is None:
                        boundsD = coords[d].create_bounds()

                    boundsD = boundsD.array

                    if shape[d] > 1:
                        tmp = numpy_empty(shape[d] + 1)
                        tmp[0:-1] = boundsD[:, 0]
                        tmp[-1] = boundsD[-1, 1]
                        boundsD = tmp

                    gridCorner[d][...] = boundsD.reshape(
                        [shape[d] + 1 if x == d else 1
                         for x in range(0, ndim)])
            # --- End: if
        # --- End: if

        # Add the mask if appropriate
        if mask is not None:
            gmask = grid.add_item(ESMF.GridItem.MASK)
            gmask[...] = 1
            gmask[mask] = 0

        return grid

    @staticmethod
    def create_field(grid, name):
        '''Create an ESMPy field for use as a source or destination field in
    regridding given an ESMPy grid and a name.

    :Parameters:

        grid: ESMF.Grid
            The ESMPy grid to use in creating the field.

        name: `str`
            The name to give the field.

    :Returns:

        `ESMF.Field`
            The resulting ESMPy field for use as a source or
            destination field in regridding.

        '''
        field = ESMF.Field(grid, name)
        return field

    def run_regridding(self, srcfield, dstfield):
        '''

        '''
        dstfield = self.regridSrc2Dst(srcfield, dstfield,
                                      zero_region=ESMF.Region.SELECT)
        return dstfield

    @staticmethod
    def concatenate_data(data_list, axis):
        '''Concatenates a list of Data objects into a single Data object along
    the specified access (see cf.Data.concatenate for details). In the
    case that the list contains only one element, that element is
    simply returned.

    :Parameters:

        data_list: `list`
            The list of data objects to concatenate.

        axis: `int`
            The axis along which to perform the concatenation.

    :Returns:

        `Data`
            The resulting single Data object.

        '''
        if len(data_list) > 1:
            data = Data.concatenate(data_list, axis=axis)
            if data.fits_in_one_chunk_in_memory(data.dtype.itemsize):
                data.varray
            return data
        else:
            assert len(data_list) == 1
            return data_list[0]

    @staticmethod
    def reconstruct_sectioned_data(sections):
        '''Expects a dictionary of Data objects with ordering information as
    keys, as output by the section method when called with a Data
    object. Returns a reconstructed cf.Data object with the sections
    in the original order.

    :Parameters:

        sections: `dict`
            The dictionary or Data objects with ordering information
            as keys.

    :Returns:

        `Data`
            The resulting reconstructed Data object.

        '''
        ndims = len(sections.keys()[0])
        for i in range(ndims - 1, -1, -1):
            keys = sorted(sections.keys())
            if i == 0:
                if keys[0][i] is None:
                    assert len(keys) == 1
                    return sections.values()[0]
                else:
                    data_list = []
                    for k in keys:
                        data_list.append(sections[k])

                    return Regrid.concatenate_data(data_list, i)

            else:
                if keys[0][i] is None:
                    pass
                else:
                    new_sections = {}
                    new_key = keys[0][:i]
                    data_list = []
                    for k in keys:
                        if k[:i] == new_key:
                            data_list.append(sections[k])
                        else:
                            new_sections[new_key] = Regrid.concatenate_data(
                                data_list, i)
                            new_key = k[:i]
                            data_list = [sections[k]]
                    # --- End: for

                    new_sections[new_key] = Regrid.concatenate_data(
                        data_list, i)
                    sections = new_sections
        # --- End: for

    @staticmethod
    def compute_mass_grid(valuefield, areafield, dofrac=False, fracfield=None,
                          uninitval=422397696.):
        '''Compute the mass of a data field.

    :Parameters:


        valuefield: ESMF.Field
            This contains data values of a field built on the cells of
            a grid.

        areafield: ESMF.Field
            This contains the areas associated with the grid cells.

        fracfield: ESMF.Field
            This contains the fractions of each cell which contributed
            to a regridding operation involving 'valuefield.

        dofrac: `bool`
            This gives the option to not use the 'fracfield'.

        uninitval: `float`
            The value uninitialised cells take.

    :Returns:

        `float`
            The mass of the data field is computed.

        '''
        mass = 0.0
        areafield.get_area()

        ind = numpy_where(valuefield.data != uninitval)

        if dofrac:
            mass = numpy_sum(
                areafield.data[ind] * valuefield.data[ind] *
                fracfield.data[ind]
            )
        else:
            mass = numpy_sum(areafield.data[ind] * valuefield.data[ind])

        return mass


# --- End: class



from os       import close
from tempfile import mkstemp
from tempfile import mkdtemp

from numpy import load    as numpy_load
from numpy import ndarray as numpy_ndarray
from numpy import save    as numpy_save

from numpy.ma import array      as numpy_ma_array
from numpy.ma import is_masked  as numpy_ma_is_masked

import cfdm

from . import abstract

from ..functions import parse_indices, get_subspace
from ..constants import CONSTANTS


class CachedArray(abstract.FileArray):
    '''A indexable N-dimensional array supporting masked values.

    The array is stored on disk in a temporary file until it is
    accessed. The directory containing the temporary file may be found
    and set with the `cf.TEMPDIR` function.

    '''
    def __init__(self, array):
        '''**Initialization**

    :Parameters:

        array: numpy array
            The array to be stored on disk in a temporary file.

    **Examples:**

    >>> f = CachedArray(numpy.array([1, 2, 3, 4, 5]))
    >>> f = CachedArray(numpy.ma.array([1, 2, 3, 4, 5]))

        '''
        super().__init__()

        # ------------------------------------------------------------
        # Use mkstemp because we want to be responsible for deleting
        # the temporary file when done with it.
        # ------------------------------------------------------------
        _partition_dir = mkdtemp(
            prefix='cf_cachedarray_', dir=CONSTANTS['TEMPDIR'])
        fd, _partition_file = mkstemp(prefix='cf_cachedarray_', suffix='.npy',
                                      dir=_partition_dir)
        close(fd)

        # The name of the temporary file storing the array
        self._set_component('_partition_dir', _partition_dir)
        self._set_component('_partition_file', _partition_file)

        # Numpy data type of the array
        self._set_component('dtype', array.dtype)

        # Tuple of the array's dimension sizes
        self._set_component('shape', array.shape)

        # Number of elements in the array
        self._set_component('size', array.size)

        # Number of dimensions in the array
        self._set_component('ndim', array.ndim)

        if numpy_ma_is_masked(array):
            # Array is a masked array. Save it as record array with
            # 'data' and 'mask' elements because this seems much
            # faster than using numpy.ma.dump.
            self._set_component('_masked_as_record', True)
            numpy_save(_partition_file, array.toflex())
        else:
            self._set_component('_masked_as_record', False)
            if hasattr(array, 'mask'):
                # Array is a masked array with no masked elements
                numpy_save(_partition_file, array.view(numpy_ndarray))
            else:
                # Array is not a masked array.
                numpy_save(_partition_file, array)

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

    Returns a numpy array.

        '''
        array = numpy_load(self._partition_file)

        indices = parse_indices(array.shape, indices)

        array = get_subspace(array, indices)

        if self._get_component('_masked_as_record'):
            # Convert a record array to a masked array
            array = numpy_ma_array(array['_data'], mask=array['_mask'],
                                   copy=False)
            array.shrink_mask()

        # Return the numpy array
        return array

    def __str__(self):
        '''x.__str__() <==> str(x)

        '''
        return '%s in %s' % (self.shape, self._partition_file)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def _partition_dir(self):
        '''TODO

        '''
        return self._get_component('_partition_dir')

    @property
    def _partition_file(self):
        '''TODO

        '''
        return self._get_component('_partition_file')


# --- End: class



import sys
import string
import numpy


_codes = {
    1: ('x', float),
    2: ('y', float),
    3: ('y_domain_lower_bound', float),
    4: ('x_domain_lower_bound', float),
    5: ('y_domain_upper_bound', float),
    6: ('x_domain_upper_bound', float),
    7: ('z_domain_lower_bound', float),
    8: ('x_domain_upper_bound', float),
    10: ('title', str),
    11: ('domain_title', str),
    12: ('x_lower_bound', float),
    13: ('x_upper_bound', float),
    14: ('y_lower_bound', float),
    15: ('y_upper_bound', float),
}


class ExtraData(dict):
    """
    Extends dictionary class with a comparison method between extra
    data for different records.
    """

    _key_to_type = dict([(key, typ) for key, typ in _codes.values()])

    def sorted_keys(self):
        k = self.keys()
        k.sort()
        return k

    _tolerances = {numpy.dtype(numpy.float32): 1e-5,
                   numpy.dtype(numpy.float64): 1e-13}

    def _cmp_floats(self, a, b, tolerance):
        if a == b:
            return 0
        delta = abs(b * tolerance)
        if a < b - delta:
            return -1
        if a > b + delta:
            return 1
        return 0

    def _cmp_float_arrays(self, avals, bvals):
        n = len(avals)
        c = cmp(n, len(bvals))
        if c != 0:
            return c
        tolerance = self._tolerances[avals.dtype]
        for i in range(n):
            c = self._cmp_floats(avals[i], bvals[i], tolerance)
            if c != 0:
                return c
        return 0

    def __cmp__(self, other):
        """Compare two extra data dictionaries returned by unpacker

        """
        if other is None:
            return 1
        ka = self.sorted_keys()
        kb = other.sorted_keys()
        c = cmp(ka, kb)
        if c != 0:
            return c
        for key in ka:
            valsa = self[key]
            valsb = other[key]
            typ = self._key_to_type[key]
            if typ == float:
                c = self._cmp_float_arrays(valsa, valsb)
            elif type == str:
                c = cmp(valsa, valsb)
            else:
                assert(False)
            if c != 0:
                return c
        return 0


class ExtraDataUnpacker:

    _int_types = {4: numpy.int32, 8: numpy.int64}
    _float_types = {4: numpy.float32, 8: numpy.float64}

    def __init__(self, raw_extra_data, word_size, byte_ordering):
        self.rdata = raw_extra_data
        self.ws = word_size
        self.itype = self._int_types[word_size]
        self.ftype = self._float_types[word_size]
        # byte_ordering is 'little_endian' or 'big_endian'
        # sys.byteorder is 'little' or 'big'
        self.is_swapped = not byte_ordering.startswith(sys.byteorder)

    def next_words(self, n):
        """
        return next n words as raw data string, and pop them off the
        front of the string
        """
        pos = n * self.ws
        rv = self.rdata[:pos]
        assert(len(rv) == pos)
        self.rdata = self.rdata[pos:]
        return rv

    def tweak_string(self, st):
        """
        undo byte-swapping of string and remove trailing NULs
        """
        if self.is_swapped:
            # concatenate backwards substrings
            st = string.join([st[pos: pos + self.ws][::-1]
                              for pos in range(0, len(st), self.ws)], "")
        while st.endswith("\x00"):
            st = st[:-1]
        return st

    def get_data(self):
        """
        get list of (key, value) for extra data
        """
        d = {}
        while self.rdata:
            i = numpy.fromstring(self.next_words(1), self.itype)[0]
            if i == 0:
                break
            ia, ib = (i / 1000, i % 1000)
            key, type = _codes[ib]
            rawvals = self.next_words(ia)
            if type == float:
                vals = numpy.fromstring(rawvals, self.ftype)
            elif type == str:
                vals = self.tweak_string(rawvals)
            d[key] = vals
        return ExtraData(d)



import datetime
import os
import time
import unittest

import cf

class FieldAncillaryTest(unittest.TestCase):
    def test_FieldAncillary(self):
        f = cf.FieldAncillary()
        self.assertTrue(f.isfieldancillary)

        _ = repr(f)
        _ = str(f)
        _ = f.dump(display=False)


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from .functions import _DEPRECATION_ERROR_FUNCTION_KWARGS
from . import Units

from .data.data import Data


def relative_vorticity(u, v, wrap=None, one_sided_at_boundary=False,
                       radius=6371229.0, cyclic=None):
    '''Calculate the relative vorticity using centred finite differences.

    The relative vorticity of wind defined on a Cartesian domain (such
    as a plane projection) is defined as

      ζcartesian = δv/δx − δu/δy

    where x and y are points on along the 'X' and 'Y' Cartesian
    dimensions respectively; and u and v denote the 'X' and 'Y'
    components of the horizontal winds.

    If the wind field field is defined on a spherical
    latitude-longitude domain then a correction factor is included:

      ζspherical = δv/δx − δu/δy + (u/a)tan(ϕ)

    where u and v denote the longitudinal and latitudinal components
    of the horizontal wind field; a is the radius of the Earth; and ϕ
    is the latitude at each point.

    The relative vorticity is calculated using centred finite
    differences (see the *one_sided_at_boundary* parameter).

    The grid may be global or limited area. If missing values are
    present then missing values will be returned at points where the
    centred finite difference could not be calculated. The boundary
    conditions may be cyclic in longitude. The non-cyclic boundaries
    may either be filled with missing values or calculated with
    off-centre finite differences.

    Reference: H.B. Bluestein, Synoptic-Dynamic Meteorology in
    Midlatitudes, 1992, Oxford Univ. Press p113-114

    :Parameters:

        u: `Field`
            A field containing the x-wind. Must be on the same grid as
            the y-wind.

        v: `Field`
            A field containing the y-wind. Must be on the same grid as
            the x-wind.

        radius: optional
            The radius of the sphere when the winds are on a spherical
            polar coordinate domain. May be any numeric scalar object
            that can be converted to a `Data` object (which includes
            numpy array and `Data` objects). By default *radius* has a
            value of 6371229.0 metres, representing the Earth's
            radius. If units are not specified then units of metres
            are assumed.

            *Parameter example:*
              Five equivalent ways to set a radius of 6371200 metres:
              ``radius=6371200``, ``radius=numpy.array(6371200)``,
              ``radius=cf.Data(6371200)``, ``radius=cf.Data(6371200,
              'm')``, ``radius=cf.Data(6371.2, 'km')``.

        wrap: `bool`, optional
            Whether the longitude is cyclic or not. By default this is
            autodetected.

        one_sided_at_boundary: `bool`, optional
            If True then if the field is not cyclic off-centre finite
            differences are calculated at the boundaries, otherwise
            missing values are used at the boundaries.

    :Returns:

        `Field`
            The relative vorticity calculated with centred finite
            differences.

    '''
    if cyclic:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            'relative_vorticity', {'cyclic': cyclic},
            "Use the 'wrap' keyword instead"
        )  # pragma: no cover

    # Get the standard names of u and v
    u_std_name = u.get_property('standard_name', None)
    v_std_name = v.get_property('standard_name', None)

    # Copy u and v
    u = u.copy()
    v = v.copy()

    # Get the X and Y coordinates
    (u_x_key, u_y_key), (u_x, u_y) = u._regrid_get_cartesian_coords(
        'u', ('X', 'Y'))
    (v_x_key, v_y_key), (v_x, v_y) = v._regrid_get_cartesian_coords(
        'v', ('X', 'Y'))

    if not u_x.equals(v_x) or not u_y.equals(v_y):
        raise ValueError('u and v must be on the same grid.')

    # Check for lat/long
    is_latlong = ((u_x.Units.islongitude and u_y.Units.islatitude) or
                  (u_x.units == 'degrees' and u_y.units == 'degrees'))

    # Check for cyclicity
    if wrap is None:
        if is_latlong:
            wrap = u.iscyclic(u_x_key)
        else:
            wrap = False
    # --- End: if

    # Find the relative vorticity
    if is_latlong:
        # Save the units of the X and Y coordinates
        x_units = u_x.Units
        y_units = u_y.Units

        # Change the units of the lat/longs to radians
        u_x.Units = Units('radians')
        u_y.Units = Units('radians')
        v_x.Units = Units('radians')
        v_y.Units = Units('radians')

        # Find cos and tan of latitude
        cos_lat = u_y.cos()
        tan_lat = u_y.tan()

        # Reshape for broadcasting
        u_shape = [1] * u.ndim
        u_y_index = u.get_data_axes().index(u_y_key)
        u_shape[u_y_index] = u_y.size

        v_shape = [1] * v.ndim
        v_y_index = v.get_data_axes().index(v_y_key)
        v_shape[v_y_index] = v_y.size

        # Calculate the correction term
        corr = u.copy()
        corr *= tan_lat.array.reshape(u_shape)

        # Calculate the derivatives
        v.derivative(v_x_key, wrap=wrap,
                     one_sided_at_boundary=one_sided_at_boundary, inplace=True)
        v.data /= cos_lat.array.reshape(v_shape)
        u.derivative(u_y_key, one_sided_at_boundary=one_sided_at_boundary,
                     inplace=True)

        radius = Data.asdata(radius).squeeze()
        radius.dtype = float
        if radius.size != 1:
            raise ValueError("Multiple radii: radius={!r}".format(radius))

        if not radius.Units:
            radius.override_units(Units('metres'), inplace=True)
        elif not radius.Units.equivalent(Units('metres')):
            raise ValueError(
                "Invalid units for radius: {!r}".format(radius.Units))

        # Calculate the relative vorticity. Do v-(u-corr) rather than
        # v-u+corr to be nice with coordinate reference corner cases.
        rv = v - (u - corr)
        rv.data /= radius

        # Convert the units of latitude and longitude to canonical units
        rv.dim('X').Units = x_units
        rv.dim('Y').Units = y_units

    else:
        v.derivative(
            v_x_key, one_sided_at_boundary=one_sided_at_boundary, inplace=True)
        u.derivative(
            u_y_key, one_sided_at_boundary=one_sided_at_boundary, inplace=True)

        rv = v - u

    # Convert the units of relative vorticity to canonical units
    rv.Units = Units('s-1')

    # Set the standard name if appropriate and delete the long_name
    if ((u_std_name == 'eastward_wind' and v_std_name == 'northward_wind') or
            (u_std_name == 'x_wind' and v_std_name == 'y_wind')):
        rv.standard_name = 'atmosphere_relative_vorticity'
    else:
        rv.del_property('standard_name', None)

    rv.del_property('long_name', None)

    return rv


def histogram(*digitized):
    '''Return the distribution of a set of variables in the form of an
    N-dimensional histogram.

    The number of dimensions of the histogram is equal to the number
    of field constructs provided by the *digitized* argument. Each
    such field construct defines a sequence of bins and provides
    indices to the bins that each value of one of the variables
    belongs. There is no upper limit to the number of dimensions of
    the histogram.

    The output histogram bins are defined by the exterior product of
    the one-dimensional bins of each digitized field construct. For
    example, if only one digitized field construct is provided then
    the histogram bins simply comprise its one-dimensional bins; if
    there are two digitized field constructs then the histogram bins
    comprise the two-dimensional matrix formed by all possible
    combinations of the two sets of one-dimensional bins; etc.

    An output value for an histogram bin is formed by counting the
    number cells for which the digitized field constructs, taken
    together, index that bin. Note that it may be the case that not
    all output bins are indexed by the digitized field constructs, and
    for these bins missing data is returned.

    The returned field construct will have a domain axis construct for
    each dimension of the histogram, with a corresponding dimension
    coordinate construct that defines the bin boundaries.

    .. versionadded:: 3.0.2

    .. seealso:: `cf.Field.bin`, `cf.Field.collapse`,
                 `cf.Field.digitize`, `cf.Field.percentile`,
                 `cf.Field.where`

    :Parameters:

        digitized: one or more `Field`
            One or more field constructs that contain digitized data
            with corresponding metadata, as would be output by
            `cf.Field.digitize`. Each field construct contains indices
            to the one-dimensional bins to which each value of an
            original field construct belongs; and there must be
            ``bin_count`` and ``bin_bounds`` properties as defined by
            the `cf.Field.digitize` method (and any of the extra
            properties defined by that method are also recommended).

            The bins defined by the ``bin_count`` and ``bin_bounds``
            properties are used to create a dimension coordinate
            construct for the output field construct.

            Each digitized field construct must be transformable so
            that its data is broadcastable to any other digitized
            field contruct's data. This is done by using the metadata
            constructs of the to create a mapping of physically
            compatible dimensions between the fields, and then
            manipulating the dimensions of the digitized field
            construct's data to ensure that broadcasting can occur.

    :Returns:

        `Field`
            The field construct containing the histogram.

    **Examples:**

    Create a one-dimensional histogram based on 10 equally-sized bins
    that exactly span the data range:

    >>> f = cf.example_field(0)
    >>> print(f)
    Field: specific_humidity (ncvar%q)
    ----------------------------------
    Data            : specific_humidity(latitude(5), longitude(8)) 1
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> print(f.array)
    [[0.007 0.034 0.003 0.014 0.018 0.037 0.024 0.029]
     [0.023 0.036 0.045 0.062 0.046 0.073 0.006 0.066]
     [0.11  0.131 0.124 0.146 0.087 0.103 0.057 0.011]
     [0.029 0.059 0.039 0.07  0.058 0.072 0.009 0.017]
     [0.006 0.036 0.019 0.035 0.018 0.037 0.034 0.013]]
    >>> indices, bins = f.digitize(10, return_bins=True)
    >>> print(indices)
    Field: long_name=Bin index to which each 'specific_humidity' value belongs (ncvar%q)
    ------------------------------------------------------------------------------------
    Data            : long_name=Bin index to which each 'specific_humidity' value belongs(latitude(5), longitude(8))
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> print(bins.array)
    [[0.003  0.0173]
     [0.0173 0.0316]
     [0.0316 0.0459]
     [0.0459 0.0602]
     [0.0602 0.0745]
     [0.0745 0.0888]
     [0.0888 0.1031]
     [0.1031 0.1174]
     [0.1174 0.1317]
     [0.1317 0.146 ]]
    >>> h = cf.histogram(indices)
    >>> rint(h)
    Field: number_of_observations
    -----------------------------
    Data            : number_of_observations(specific_humidity(10)) 1
    Cell methods    : latitude: longitude: point
    Dimension coords: specific_humidity(10) = [0.01015, ..., 0.13885] 1
    >>> print(h.array)
    [9 7 9 4 5 1 1 1 2 1]
    >>> print(h.coordinate('specific_humidity').bounds.array)
    [[0.003  0.0173]
     [0.0173 0.0316]
     [0.0316 0.0459]
     [0.0459 0.0602]
     [0.0602 0.0745]
     [0.0745 0.0888]
     [0.0888 0.1031]
     [0.1031 0.1174]
     [0.1174 0.1317]
     [0.1317 0.146 ]]


    Create a two-dimensional histogram based on specific humidity and
    temperature bins. The temperature bins in this example are derived
    from a dummy temperature field construct with the same shape as
    the specific humidity field construct already in use:

    >>> g = f.copy()
    >>> g.standard_name = 'air_temperature'
    >>> import numpy
    >>> g[...] = numpy.random.normal(loc=290, scale=10, size=40).reshape(5, 8)
    >>> g.overide_units('K', inplace=True)
    >>> print(g)
    Field: air_temperature (ncvar%q)
    --------------------------------
    Data            : air_temperature(latitude(5), longitude(8)) K
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]

    >>> indices_t = g.digitize(5)
    >>> h = cf.histogram(indices, indices_t)
    >>> print(h)
    Field: number_of_observations
    -----------------------------
    Data            : number_of_observations(air_temperature(5), specific_humidity(10)) 1
    Cell methods    : latitude: longitude: point
    Dimension coords: air_temperature(5) = [281.1054839143287, ..., 313.9741786365939] K
                    : specific_humidity(10) = [0.01015, ..., 0.13885] 1
    >>> print(h.array)
    [[2  1  5  3  2 -- -- -- -- --]
     [1  1  2 --  1 --  1  1 -- --]
     [4  4  2  1  1  1 -- --  1  1]
     [1  1 -- --  1 -- -- --  1 --]
     [1 -- -- -- -- -- -- -- -- --]]
    >>> h.sum()
    <CF Data(): 40 1>

    '''
    if not digitized:
        raise ValueError(
            "Must provide at least one 'digitized' field construct")

    f = digitized[0].copy()
    f.clear_properties()

    return f.bin('sample_size', digitized=digitized)



from copy import deepcopy

from ..functions import ATOL, RTOL

from ..query import Query
from ..units import Units

from ..functions import (_DEPRECATION_ERROR_METHOD,
                         _DEPRECATION_ERROR)


class Properties:
    '''Mixin class for a container of descriptive properties.

    .. versionadded:: 3.0.0

    '''
    _special_properties = ()

    # ----------------------------------------------------------------
    # Private attributes
    # ----------------------------------------------------------------
    @property
    def _ATOL(self):
        '''Return the tolerance on absolute differences between real numbers,
    as returned by the `cf.ATOL` function.

    This is used by, for example, the `_equals` method.

        '''
        return ATOL()

    @property
    def _RTOL(self):
        '''Return the tolerance on relative differences between real numbers,
    as returned by the `cf.RTOL` function.

    This is used by, for example, the `_equals` method.

        '''
        return RTOL()

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _matching_values(self, value0, value1, units=False):
        '''Whether two values match.

    The definition of "match" depends on the types of *value0* and
    *value1*. See the code for the details.

    :Parameters:

        value0:
            The first value to be matched.

        value1:
            The second value to be matched.

        units: `bool`, optional

    :Returns:

        `bool`
            Whether or not the two values match.

        '''
        if value1 is None:
            return False

        if isinstance(value0, Query):
            return bool(value0.evaluate(value1))  # TODO vectors

        try:
            # value0 is a re.compile object
            return bool(value0.search(value1))
        except (AttributeError, TypeError):
            if units and isinstance(value0, str):
                return Units(value0).equals(Units(value1))

            return self._equals(value1, value0)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def id(self):
        '''A canonical identity.

    The `id` attribute can be used to unambiguously identify
    constructs in the absence of a `standard_name` property.

    Note that `id` is not a CF property and so is not read from, or
    written to, datasets.

    **Examples:**

    >>> f.id = 'um01002'
    >>> f.id
    'um01002'
    >>> del f.id

        '''
        try:
            return self._custom['id']
        except KeyError:
            raise AttributeError("{} doesn't have attribute 'id'".format(
                self.__class__.__name__))

    @id.setter
    def id(self, value):   self._custom['id'] = value

    @id.deleter
    def id(self):
        try:
            del self._custom['id']
        except KeyError:
            raise AttributeError("{} doesn't have attribute 'id'".format(
                self.__class__.__name__))

    # ----------------------------------------------------------------
    # CF properties
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''The calendar CF property.

    The calendar used for encoding time data. See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.calendar = 'noleap'
    >>> f.calendar
    'noleap'
    >>> del f.calendar

    >>> f.set_property('calendar', 'proleptic_gregorian')
    >>> f.has_property('calendar')
    True
    >>> f.get_property('calendar')
    'proleptic_gregorian'
    >>> f.del_property('calendar')

        '''
        return self.get_property('calendar', default=AttributeError())

    @calendar.setter
    def calendar(self, value): self.set_property('calendar', value)

    @calendar.deleter
    def calendar(self):
        self.del_property('calendar', default=AttributeError())

    @property
    def comment(self):
        '''The comment CF property.

    Miscellaneous information about the data or methods used to
    produce it. See http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.comment = 'This simulation was done on an HP-35 calculator'
    >>> f.comment
    'This simulation was done on an HP-35 calculator'
    >>> del f.comment

    >>> f.set_property('comment', 'a comment')
    >>> f.has_property('comment')
    True
    >>> f.get_property('comment')
    'a comment'
    >>> f.del_property('comment')

        '''
        return self.get_property('comment', default=AttributeError())

    @comment.setter
    def comment(self, value): self.set_property('comment', value)

    @comment.deleter
    def comment(self):
        self.del_property('comment', default=AttributeError())

    @property
    def history(self):
        '''The history CF property.

    A list of the applications that have modified the original
    data. See http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.history = 'created on 2012/10/01'
    >>> f.history
    'created on 2012/10/01'
    >>> del f.history

    >>> f.set_property('history', 'created on 2012/10/01')
    >>> f.has_property('history')
    True
    >>> f.get_property('history')
    'created on 2012/10/01'
    >>> f.del_property('history')

        '''
        return self.get_property('history', default=AttributeError())

    @history.setter
    def history(self, value):
        self.set_property('history', value)

    @history.deleter
    def history(self):
        self.del_property('history', default=AttributeError())

    @property
    def leap_month(self):
        '''The leap_month CF property.

    Specifies which month is lengthened by a day in leap years for a
    user defined calendar. See http://cfconventions.org/latest.html
    for details.

    **Examples:**

    >>> f.leap_month = 2
    >>> f.leap_month
    2
    >>> del f.leap_month

    >>> f.set_property('leap_month', 11)
    >>> f.has_property('leap_month')
    True
    >>> f.get_property('leap_month')
    11
    >>> f.del_property('leap_month')

        '''
        return self.get_property('leap_month', default=AttributeError())

    @leap_month.setter
    def leap_month(self, value):
        self.set_property('leap_month', value)

    @leap_month.deleter
    def leap_month(self):
        self.del_property('leap_month', default=AttributeError())

    @property
    def leap_year(self):
        '''The leap_year CF property.

    Provides an example of a leap year for a user defined calendar. It
    is assumed that all years that differ from this year by a multiple
    of four are also leap years. See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.leap_year = 1984
    >>> f.leap_year
    1984
    >>> del f.leap_year

    >>> f.set_property('leap_year', 1984)
    >>> f.has_property('leap_year')
    True
    >>> f.get_property('leap_year')
    1984
    >>> f.del_property('leap_year')

        '''
        return self.get_property('leap_year', default=AttributeError())

    @leap_year.setter
    def leap_year(self, value):
        self.set_property('leap_year', value)

    @leap_year.deleter
    def leap_year(self):
        self.del_property('leap_year', default=AttributeError())

    @property
    def long_name(self):
        '''The long_name CF property.

    A descriptive name that indicates a nature of the data. This name
    is not standardized. See http://cfconventions.org/latest.html for
    details.

    **Examples:**

    >>> f.long_name = 'zonal_wind'
    >>> f.long_name
    'zonal_wind'
    >>> del f.long_name

    >>> f.set_property('long_name', 'surface air temperature')
    >>> f.has_property('long_name')
    True
    >>> f.get_property('long_name')
    'surface air temperature'
    >>> f.del_property('long_name')

        '''
        return self.get_property('long_name', default=AttributeError())

    @long_name.setter
    def long_name(self, value):
        self.set_property('long_name', value)

    @long_name.deleter
    def long_name(self):
        self.del_property('long_name', default=AttributeError())

    @property
    def month_lengths(self):
        '''The month_lengths CF property.

    Specifies the length of each month in a non-leap year for a user
    defined calendar. See http://cfconventions.org/latest.html for
    details.

    Stored as a tuple but may be set as any array-like object.

    **Examples:**

    >>> f.month_lengths = numpy.array(
    ...     [34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34])
    >>> f.month_lengths
    (34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34)
    >>> del f.month_lengths

    >>> f.set_property('month_lengths',
    ...                [34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34])
    >>> f.has_property('month_lengths')
    True
    >>> f.get_property('month_lengths')
    (34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34)
    >>> f.del_property('month_lengths')

        '''
        return self.get_property('month_lengths', default=AttributeError())

    @month_lengths.setter
    def month_lengths(self, value):
        self.set_property('month_lengths', tuple(value))

    @month_lengths.deleter
    def month_lengths(self):
        self.del_property('month_lengths', default=AttributeError())

    @property
    def standard_name(self):
        '''The standard_name CF property.

    A standard name that references a description of a data in the
    standard name table
    (http://cfconventions.org/standard-names.html). See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.standard_name = 'time'
    >>> f.standard_name
    'time'
    >>> del f.standard_name

    >>> f.set_property('standard_name', 'time')
    >>> f.has_property('standard_name')
    True
    >>> f.get_property('standard_name')
    'time'
    >>> f.del_property('standard_name')

        '''
        return self.get_property('standard_name', default=AttributeError())

    @standard_name.setter
    def standard_name(self, value): self.set_property('standard_name', value)

    @standard_name.deleter
    def standard_name(self):
        self.del_property('standard_name', default=AttributeError())

    @property
    def units(self):
        '''The units CF property.

    The units of the data. The value of the `units` property is a
    string that can be recognized by UNIDATA's Udunits package
    (http://www.unidata.ucar.edu/software/udunits). See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.units = 'K'
    >>> f.units
    'K'
    >>> del f.units

    >>> f.set_property('units', 'm.s-1')
    >>> f.has_property('units')
    True
    >>> f.get_property('units')
    'm.s-1'
    >>> f.del_property('units')

        '''
        return self.get_property('units', default=AttributeError())

    @units.setter
    def units(self, value): self.set_property('units', value)

    @units.deleter
    def units(self):
        self.del_property('units', default=AttributeError())

    @property
    def valid_max(self):
        '''The valid_max CF property.

    The largest valid value of the data. See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.valid_max = 100.0
    >>> f.valid_max
    100.0
    >>> del f.valid_max

    >>> f.set_property('valid_max', 100.0)
    >>> f.has_property('valid_max')
    True
    >>> f.get_property('valid_max')
    100.0
    >>> f.del_property('valid_max')

        '''
        return self.get_property('valid_max', default=AttributeError())

    @valid_max.setter
    def valid_max(self, value): self.set_property('valid_max', value)

    @valid_max.deleter
    def valid_max(self):
        self.del_property('valid_max', default=AttributeError())

    @property
    def valid_min(self):
        '''The valid_min CF property.

    The smallest valid value of the data. See
    http://cfconventions.org/latest.html for details.

    **Examples:**

    >>> f.valid_min = 8.0
    >>> f.valid_min
    8.0
    >>> del f.valid_min

    >>> f.set_property('valid_min', 8.0)
    >>> f.has_property('valid_min')
    True
    >>> f.get_property('valid_min')
    8.0
    >>> f.del_property('valid_min')

        '''
        return self.get_property('valid_min', default=AttributeError())

    @valid_min.setter
    def valid_min(self, value): self.set_property('valid_min', value)

    @valid_min.deleter
    def valid_min(self):
        self.del_property('valid_min', default=AttributeError())

    @property
    def valid_range(self):
        '''The valid_range CF property.

    The smallest and largest valid values the data. See
    http://cfconventions.org/latest.html for details.

    Stored as a tuple but may be set as any array-like object.

    **Examples:**

    >>> f.valid_range = numpy.array([100., 400.])
    >>> f.valid_range
    (100.0, 400.0)
    >>> del f.valid_range

    >>> f.set_property('valid_range', [100.0, 400.0])
    >>> f.has_property('valid_range')
    True
    >>> f.get_property('valid_range')
    (100.0, 400.0)
    >>> f.del_property('valid_range')

        '''
        return tuple(
            self.get_property('valid_range'), default=AttributeError())

    @valid_range.setter
    def valid_range(self, value):
        self.set_property('valid_range', tuple(value))

    @valid_range.deleter
    def valid_range(self):
        self.del_property('valid_range', default=AttributeError())

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def get_property(self, prop, default=ValueError()):
        '''Get a CF property.

    .. versionadded:: 3.0.0

    .. seealso:: `clear_properties`, `del_property`, `has_property`,
                 `properties`, `set_property`

    :Parameters:

        prop: `str`
            The name of the CF property.

            *Parameter example:*
              ``prop='long_name'``

        default: optional
            Return the value of the *default* parameter if the
            property does not exist. If set to an `Exception` instance
            then it will be raised instead.

    :Returns:

            The value of the named property or the default value, if
            set.

    **Examples:**


    >>> f.set_property('project', 'CMIP7')
    >>> f.has_property('project')
    True
    >>> f.get_property('project')
    'CMIP7'
    >>> f.del_property('project')
    'CMIP7'
    >>> f.has_property('project')
    False
    >>> print(f.del_property('project', None))
    None
    >>> print(f.get_property('project', None))
    None

        '''
        # Get a special property
        if prop in self._special_properties:
            try:
                return getattr(self, prop)
            except AttributeError as err:
                return self._default(default, err)
        # --- End: if

        # Still here? Then get a non-special property
        return super().get_property(prop, default=default)

    def has_property(self, prop):
        '''Whether a property has been set.

    .. versionadded:: 3.0.0

    .. seealso:: `clear_properties`, `del_property`, `get_property`,
                 `properties`, `set_property`

    :Parameters:

        prop: `str`
            The name of the property.

            *Parameter example:*
               ``prop='long_name'``

    :Returns:

         `bool`
            `True` if the property has been set, otherwise `False`.

    **Examples:**

    >>> f.set_property('project', 'CMIP7')
    >>> f.has_property('project')
    True
    >>> f.get_property('project')
    'CMIP7'
    >>> f.del_property('project')
    'CMIP7'
    >>> f.has_property('project')
    False
    >>> print(f.del_property('project', None))
    None
    >>> print(f.get_property('project', None))
    None

        '''
        # Get a special property
        if prop in self._special_properties:
            return hasattr(self, prop)

        # Still here? Then get a non-special property
        return super().has_property(prop)

    def del_property(self, prop, default=ValueError()):
        '''Remove a property.

    .. versionadded:: 3.0.0

    .. seealso:: `clear_properties`, `get_property`, `has_property`,
                 `properties`, `set_property`

    :Parameters:

        prop: `str`
            The name of the property.

            *Parameter example:*
               ``prop='long_name'``

        default: optional
            Return the value of the *default* parameter if the
            property does not exist. If set to an `Exception` instance
            then it will be raised instead.

    :Returns:

            The removed property.

    **Examples:**

    >>> f.set_property('project', 'CMIP7')
    >>> f.has_property('project')
    True
    >>> f.get_property('project')
    'CMIP7'
    >>> f.del_property('project')
    'CMIP7'
    >>> f.has_property('project')
    False
    >>> print(f.del_property('project', None))
    None
    >>> print(f.get_property('project', None))
    None

        '''
        # Get a special attribute
        if prop in self._special_properties:
            try:
                out = getattr(self, prop)
            except AttributeError as err:
                return self._default(default, err)
            else:
                delattr(self, prop)
                return out
        # --- End: if

        # Still here? Then del a non-special attribute
        return super().del_property(prop, default=default)

    def match_by_identity(self, *identities):
        '''Whether or not the construct identity satisfies conditions.

    .. versionadded:: 3.0.0

    .. seealso:: `match`, `match_by_property`, `match_by_ncvar`

    :Parameters:

        identities: optional
            Define one or more conditions on the identities.

            A construct identity is specified by a string
            (e.g. ``'latitude'``, ``'long_name=time', ``'ncvar%lat'``,
            etc.); a `Query` object (e.g. ``cf.eq('longitude')``); or
            a compiled regular expression
            (e.g. ``re.compile('^atmosphere')``) that is compared with
            the construct's identities via `re.search`.

            A construct has a number of identities, and the condition
            is satisfied if any of the construct's identities, as
            returned by the `identities` method, equals the condition
            value. A construct's identities are those returned by its
            `!identities` method. In the following example, the
            construct ``x`` has six identities:

               >>> x.identities()
               ['time',
                'long_name=Time',
                'foo=bar',
                'standard_name=time',
                'ncvar%t',
                'T']

    :Returns:

        `bool`
            Whether or not at least one of the conditions are met.

    **Examples:**

    >>> f.match_by_identity('time')

    >>> f.match_by_identity(re.compile('^air'))

    >>> f.match_by_identity('air_pressure', 'air_temperature')

    >>> f.match_by_identity('ncvar%t')

        '''
        if not identities:
            return True

        self_identities = self.identities()

        ok = False
        for value0 in identities:
            for value1 in self_identities:
                ok = self._matching_values(value0, value1)
                if ok:
                    break
            # --- End: for

            if ok:
                break
        # --- End: for

        return ok

    def match_by_ncvar(self, *ncvars):
        '''Whether or not the netCDF variable name satisfies conditions.

    .. versionadded:: 3.0.0

    .. seealso:: `match`, `match_by_identity`, `match_by_property`

    :Parameters:

        ncvars: optional
            Define one or more conditions on the netCDF variable name.

            A netCDF variable name is specified by a string
            (e.g. ``'lat'``), etc.); a `Query` object
            (e.g. ``cf.eq('lon')``); or a compiled regular expression
            (e.g. ``re.compile('^t')``) that is compared with the
            construct's netCDF variable name via `re.search`.

            The condition is satisfied if the netCDF variable name, as
            returned by the `nc_get_variable` method, exists and
            equals the condition value.

    :Returns:

        `bool`
            Whether or not at least one of the conditions are met.

    **Examples:**

    >>> f.nc_get_variable()
    'time'
    >>> f.match_by_ncvar('time')
    True
    >>> f.match_by_ncvar('t', 'time')
    True
    >>> f.match_by_ncvar('t')
    False
    >>> f.match_by_ncvar()
    True
    >>> import re
    >>> f.match_by_ncvar(re.compile('^t'))
    True

        '''
        if not ncvars:
            return True

        try:
            ncvar = self.nc_get_variable(None)
        except AttributeError:
            ncvar = None

        if ncvar is None:
            return False

        ok = False
        for value0 in ncvars:
            ok = self._matching_values(value0, ncvar)
            if ok:
                break
        # --- End: for

        return ok

    def match_by_property(self, *mode, **properties):
        '''Whether or not properties satisfy conditions.

    .. versionadded:: 3.0.0

    .. seealso:: `match`, `match_by_identity`, `match_by_ncvar`

    :Parameters:

        mode: optional
            Define the behaviour when multiple conditions are
            provided.

            By default (or if the *mode* parameter is ``'and'``) the
            match is `True` if the field construct satisfies all of
            the given conditions, but if the *mode* parameter is
            ``'or'`` then the match is `True` when at least one of the
            conditions is satisfied.

        properties: optional
            Define conditions on properties.

            Properties are identified by the name of a keyword
            parameter.

            The keyword parameter value defines a condition to be
            applied to the property, and is either a string
            (e.g. ``'latitude'``); a `Query` object
            (e.g. ``cf.eq('longitude')``); or a compiled regular
            expression (e.g. ``re.compile('^atmosphere')``) that is
            compared with the property value via `re.search`.

            If the value is `None` then the condition is one of
            existence and is satisfied if the property exists,
            regardless of its value. Otherwise the condition is
            satisfied if the property equals the value.

            *Parameter example:*
              To see if there is a 'standard_name' property of 'time':
              ``standard_name='time'``.

            *Parameter example:*
              To see if there is a 'valid_min' property of that is
              less than -999: ``valid_min=cf.lt('-999)``.

            *Parameter example:*
              To see if there is a 'long_name' property with any
              value: ``long_name=None``.

    :Returns:

        `bool`
            Whether or not the conditions are met.

    **Examples:**

    >>> f.match_by_property(standard_name='longitude')

    >>> f.match_by_property(
    ...     standard_name='longitude', foo='cf.set(['bar', 'not_bar'])

    >>> f.match_by_property(long_name=re.compile('^lon'))

        '''
        _or = False

        # Parse mode
        if mode:
            if len(mode) > 1:
                raise ValueError(
                    "Can provide at most one positional argument")

            x = mode[0]
            if x == 'or':
                _or = True
            elif x != 'and':
                raise ValueError(
                    "Positional argument, if provided, must one of 'or', "
                    "'and'"
                )
        # --- End: if

        if not properties:
            return True

        self_properties = self.properties()

        ok = True
        for name, value0 in properties.items():
            if value0 is None:
                ok = (name in self_properties)
            else:
                value1 = self_properties.get(name)
                ok = self._matching_values(value0, value1,
                                           units=(name == 'units'))

            if _or:
                if ok:
                    break
            elif not ok:
                break
        # --- End: for

        return ok

    def properties(self):
        '''Return all properties.

    .. seealso:: `clear_properties`, `get_property`, `has_property`
                 `set_properties`

    :Returns:

        `dict`
            The properties.

    **Examples:**

    >>> f.properties()
    {}
    >>> f.set_properties(
    ...     {'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.clear_properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.properties()
    {}

        '''
        out = super().properties()

        for prop in self._special_properties:
            value = getattr(self, prop, None)
            if value is None:
                out.pop(prop, None)
            else:
                out[prop] = value
        # --- End: for

        return out

    def set_properties(self, properties, copy=True):
        '''Set properties.

    .. versionadded:: 3.0.0

    .. seealso:: `clear_properties`, `properties`, `set_property`

    :Parameters:

        properties: `dict`
            Store the properties from the dictionary supplied.

            *Parameter example:*
              ``properties={'standard_name': 'altitude', 'foo': 'bar'}``

        copy: `bool`, optional
            If `False` then any property values provided by the
            *properties* parameter are not copied before insertion. By
            default they are deep copied.

    :Returns:

        `None`

    **Examples:**

    >>> f.properties()
    {}
    >>> f.set_properties(
    ...     {'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.clear_properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.properties()
    {}

        '''
        super().set_properties(properties, copy=copy)

        for prop in self._special_properties:
            value = properties.get(prop)
            if value is not None:
                setattr(self, prop, value)

    def set_property(self, prop, value, copy=True):
        '''Set a property.

    .. versionadded:: 3.0.0

    .. seealso:: `clear_properties`, `del_property`, `get_property`,
                 `has_property`, `properties`

    :Parameters:

        prop: `str`
            The name of the property to be set.

        value:
            The value for the property.

        copy: `bool`, optional
            If `True` then set a deep copy of *value*.

    :Returns:

         `None`

    **Examples:**

    >>> f.set_property('project', 'CMIP7')
    >>> f.has_property('project')
    True
    >>> f.get_property('project')
    'CMIP7'
    >>> f.del_property('project')
    'CMIP7'
    >>> f.has_property('project')
    False
    >>> print(f.del_property('project', None))
    None
    >>> print(f.get_property('project', None))
    None

        '''
        # Get a special property
        if prop in self._special_properties:
            if copy:
                value = deepcopy(value)

            setattr(self, prop, value)
            return

        # Still here? Then set a non-special property
        return super().set_property(prop, value, copy=copy)

    # ----------------------------------------------------------------
    # Aliases
    # ----------------------------------------------------------------
    def match(self, *identities, **kwargs):
        '''Alias for `match_by_identity`.

        '''
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self, 'match', kwargs, "Use 'match_by_*' methods instead."
            )  # pragma: no cover

        if identities and isinstance(identities[0], (list, tuple, set)):
            _DEPRECATION_ERROR(
                "Use of a {!r} for identities has been deprecated. Use the "
                "* operator to unpack the arguments instead.".format(
                    identities[0].__class__.__name__)
            )  # pragma: no cover

        for i in identities:
            if isinstance(i, dict):
                _DEPRECATION_ERROR_DICT(
                    "Use 'match_by_*' methods instead.")  # pragma: no cover
        # --- End: for

        return self.match_by_identity(*identities)

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def setprop(self, *args, **kwargs):
        '''Delete a CF property.

    Deprecated at version 3.0.0. Use method 'set_property' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'setprop', "Use method 'set_property' instead"
        )  # pragma: no cover

    def delprop(self, prop):
        '''Delete a CF property.

    Deprecated at version 3.0.0. Use method 'del_property' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'delprop', "Use method 'del_property' instead"
        )  # pragma: no cover

    def hasprop(self, prop):
        '''Whether a property has been set.

    Deprecated at version 3.0.0. Use method 'has_property' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'hasprop', "Use method 'has_property' instead"
        )  # pragma: no cover

    def getprop(self, prop):
        '''Get a property.

    Deprecated at version 3.0.0. Use method 'get_property' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'getprop', "Use method 'get_property' instead"
        )  # pragma: no cover


# --- End: class
